//
//  DeclarationCollector.swift
//  SwiftMetaModelGenerator
//
//  Created by Tobias on 23.06.21.
//

import Foundation
import SwiftSyntax

class ClassDeclaration {
    let name: String
    let parents: [String]
    
    let variables: [Variable]
    var needsMethodOverride: Bool
    
    init(name: String, parents: [String], variables: [Variable]) {
        self.name = name
        self.parents = parents
        self.variables = variables
        self.needsMethodOverride = false
    }
}

class Variable {
    let name: String
    let type: String?
    let optional: Bool?
    let constant: Bool
    let privateVisibility: Bool
    let staticModifier: Bool
    let hasAccessor: Bool
    
    init(name: String, type: String?, optional: Bool?, constant: Bool, privateVisibility: Bool, staticModifier: Bool, hasAccessor: Bool) {
        self.name = name
        self.type = type
        self.optional = optional
        self.constant = constant
        self.privateVisibility = privateVisibility
        self.staticModifier = staticModifier
        self.hasAccessor = hasAccessor
    }
}

class DeclarationCollector: SyntaxVisitor {
    
    private(set) var classes: [ClassDeclaration] = []
    private var currentVariables: [Variable] = []
    
    override func visitPost(_ node: ClassDeclSyntax) {
        let name = node.identifier.text.trimmingCharacters(in: .whitespacesAndNewlines)
        let inherences = node.inheritanceClause?.inheritedTypeCollection.map({ $0.typeName.description.trimmingCharacters(in: .whitespacesAndNewlines) }) ?? []
        
        classes.append(ClassDeclaration(name: name, parents: inherences, variables: currentVariables))
        currentVariables = []
    }
    
    override func visit(_ node: VariableDeclSyntax) -> SyntaxVisitorContinueKind {
        for child: PatternBindingSyntax in node.bindings {
            if let parent = child.context as? VariableDeclSyntax, parent.context is ClassDeclSyntax {
                let name = child.pattern.description.trimmingCharacters(in: .whitespacesAndNewlines)
                let type = child.typeAnnotation?.type.description.trimmingCharacters(in: .whitespacesAndNewlines)
                let constant = parent.letOrVarKeyword.text.trimmingCharacters(in: .whitespacesAndNewlines) == "let"
                
                let privateVisibility = parent.modifiers?.map({ $0.name.text }).contains("private") ?? false
                let staticModifier = parent.modifiers?.map({ $0.name.text }).contains("static") ?? false
                let hasAccessor = child.accessor != nil
                
                if type == nil {
                    print("Fail to observe type for \(name)")
                }
                
                currentVariables.append(Variable(name: name, type: type?.replacingOccurrences(of: "?", with: ""), optional: type?.hasSuffix("?") , constant: constant, privateVisibility: privateVisibility, staticModifier: staticModifier, hasAccessor: hasAccessor))
            }
        }
        return .skipChildren
    }
}


extension SyntaxProtocol {
    var context: DeclSyntaxProtocol? {
        for case let node? in sequence(first: parent, next: { $0?.parent }) {
            guard let declaration = node.asProtocol(DeclSyntaxProtocol.self) else { continue }
            return declaration
        }
        
        return nil
    }
}
